import logger from "./logger";
import {InternalMessage} from "./constants";

type InitialUserData = {
    sipLogin: string,
    sipVisibleName: string,
    pmi: string
}

type OnMessageCallback = (arg0: string, arg1: Array<InternalMessage>) => void;

export class Connection {
    private ws: WebSocket;
    private onError: Function;
    private onClose: Function;
    private onMessage: OnMessageCallback;
    private connected: boolean;
    private connectionTimeout: number;

    public constructor(onMessage: OnMessageCallback, onError: Function, onClose: Function) {
        this.connected = false;
        this.onMessage = onMessage;
        this.onError = onError;
        this.onClose = onClose;
    }

    public connect(options: {
        url: string,
        appName: string,
        timeout: number,
        custom: object | null
    }): Promise<InitialUserData> {
        this.ws = new WebSocket(options.url);
        this.ws.binaryType = 'arraybuffer';
        const that = this;
        return new Promise(function(resolve, reject) {
            if (options.timeout !== undefined && options.timeout > 0) {
                that.connectionTimeout = window.setTimeout(function () {
                    if (that.ws.readyState === 0) {
                        console.log("WS connection timeout");
                        that.ws.close();
                    }
                }, options.timeout);
            }
            that.ws.onerror = function (e) {
                if (that.connected) {
                    that.onError(e);
                } else {
                    reject(e);
                }
            };
            that.ws.onclose = function (e) {
                if (that.connected) {
                    that.onClose(e);
                } else {
                    reject(e);
                }
            };
            that.ws.onopen = function () {
                clearTimeout(that.connectionTimeout);
                let cConfig = {
                    appKey: options.appName,
                    mediaProviders: ["WebRTC"],
                    keepAlive: true,
                    authToken: null,
                    clientVersion: "0.5.28",
                    clientOSVersion: window.navigator.appVersion,
                    clientBrowserVersion: window.navigator.userAgent,
                    custom: options.custom
                };
                //connect to REST App
                that.send("connection", cConfig);
            };
            that.ws.onmessage = function (event: MessageEvent<Blob | string>) {
                let name;
                let msg: Array<InternalMessage>;
                if (event.data instanceof Blob) {
                    name = "binaryData";
                } else {
                    let obj = JSON.parse(event.data) as {
                        message: string,
                        data: Array<InternalMessage>
                    };
                    name = obj.message;
                    msg = obj.data;
                }
                switch (name) {
                    case "ping":
                        that.send("pong", null);
                        break;
                    case "getUserData":
                        that.connected = true;
                        resolve(msg[0] as unknown as InitialUserData);
                        break;
                    default:
                        that.onMessage(name, msg);
                }
            };
        });
    }

    public send(message: string, data: any) {
        logger.log("==>", message, data);
        if (this.ws) {
            this.ws.send(JSON.stringify({
                message: message,
                data: [data]
            }));
        }
    };

    public close() {
        if (this.ws) {
            this.ws.close();
        }
    }
}