import {v4 as uuidv4} from 'uuid';
import {Room} from "./room";
import {Connection} from "./connection";
import promises from "./promises";
import {InternalApi, ParticipantRole, WaitingRoomUpdate} from "./constants";

export class RoomExtended extends Room {

    public constructor(connection: Connection, pc: RTCPeerConnection, name: string, pin: string) {
       super(connection, pc, name, pin);
    }

    public sendControlMessage(to: string, msg: string, broadcast: boolean) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.SEND_CONTROL_MESSAGE, {
                broadcast: broadcast,
                from: "",
                to: to,
                body: msg,
                internalMessageId: id
            });
        });
    };

    public authorizeWaitingList(userId: string, authorized: boolean) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.AUTHORIZE_WAITING_LIST, {
                name: self._name,
                userId: userId,
                authorized: authorized,
                internalMessageId: id
            });
        });
    };

    public moveToWaitingRoom(nickname: string) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.MOVE_TO_WAITING_ROOM, {
                roomName: self._name,
                nickname: nickname,
                internalMessageId: id
            });
        });
    };

    public configureWaitingRoom(enabled: boolean) {
        const self = this;
        return new Promise<WaitingRoomUpdate>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.CONFIGURE_WAITING_ROOM, {
                name: self._name,
                enabled: enabled,
                internalMessageId: id
            });
        });
    };

    public assignRole(nickname: string, role: ParticipantRole) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.ASSIGN_ROLE, {
                roomName: self._name,
                nickname: nickname,
                role: role,
                internalMessageId: id
            });
        });
    };

    public subscribeToWaitingParticipant(nickname: string) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.SUBSCRIBE_TO_WAITING_PARTICIPANT, {
                roomName: self._name,
                nickname: nickname,
                internalMessageId: id
            });
        });
    };

    public unsubscribeFromWaitingParticipant(nickname: string) {
        const self = this;
        return new Promise<void>((resolve, reject) => {
            const id = uuidv4();
            promises.add(id, resolve, reject);
            self.connection.send(InternalApi.UNSUBSCRIBE_FROM_WAITING_PARTICIPANT, {
                roomName: self._name,
                nickname: nickname,
                internalMessageId: id
            });
        });
    };
}
