import {v4 as uuidv4} from 'uuid';
import promises from "./promises";
import {Connection} from "./connection";
import {
    Calendar, ChannelSendPolicy, ChannelType,
    Chat,
    ChatLoadedEvent,
    ChatMap,
    ChatsEvent,
    ContactInviteEvent,
    ContactRemovedEvent,
    ContactUpdateEvent,
    InternalApi,
    InternalMessage,
    Invite, LeftRoom,
    Message, MessageAttachment,
    MessageStatus,
    MessageStatusEvent,
    NewChatEvent,
    OperationFailedEvent, Operations,
    PublicChannelsEvent,
    RemovedChatEvent,
    RoomEvent,
    SfuEvent,
    SfuMessageEvent,
    State,
    UpdateChatEvent,
    User,
    UserCalendarEvent,
    UserListEvent,
    UserSpecificChatInfo
} from "./constants";
import {Notifier} from "./notifier";
import {RoomExtended} from "./room-extended";

type NotifyUnion = InternalMessage | Message | MessageStatus | Array<User> | Calendar | UserSpecificChatInfo | Invite | User | ChatMap | Chat;

export class SfuExtended {

    #connection: Connection;
    #_user: {
        username: string,
        nickname: string,
        pmi: string
    }
    #_server: string;
    #_state: State = State.NEW;
    #rooms: {[key: string]: RoomExtended} = {};
    //TODO(naz): Provide union instead of InternalMessage
    #notifier: Notifier<SfuEvent, NotifyUnion> = new Notifier<SfuEvent, NotifyUnion>();

    public connect(options: {
        url: string,
        username: string,
        password: string,
        nickname: string,
        timeout?: number
    }) {
        if (!options) {
            throw new TypeError("No options provided");
        }
        const connectionConfig = {
            url: options.url,
            appName: InternalApi.Z_APP,
            timeout: options.timeout ? options.timeout : 10000,
            custom: {
                username: options.username,
                password: options.password,
                nickname: options.nickname
            }
        };
        const self = this;
        this.#createConnection();
        this.#_server = new URL(options.url).hostname;
        return new Promise<{
            username: string,
            nickname: string,
            pmi: string
        }>(async (resolve, reject) => {
            try {
                const userCredentials = await this.#connection.connect(connectionConfig);
                self.#_user = {
                    username: userCredentials.sipLogin,
                    nickname: userCredentials.sipVisibleName,
                    pmi: userCredentials.pmi
                }
                self.#_state = State.AUTHENTICATED;
                self.#notifier.notify(SfuEvent.CONNECTED);
                resolve(self.#_user);
            } catch (e) {
                self.#_state = State.FAILED;
                this.#notifier.notify(SfuEvent.FAILED, e);
                reject(e);
            }
        });
    };

    #createConnection() {
        const self = this;
        this.#connection = new Connection(
            (name: string, data: InternalMessage[]) => {
                switch (name) {
                    case InternalApi.DEFAULT_METHOD:
                        //filter messages
                        //TODO(naz): refactor this part
                        if (data[0].type === InternalApi.MESSAGE) {
                            const message = (data[0] as SfuMessageEvent).message;
                            this.#notifier.notify(SfuEvent.MESSAGE, message);
                        } else if (data[0].type === InternalApi.MESSAGE_STATE) {
                            const messageState = data[0] as MessageStatusEvent;
                            if (!promises.resolve(data[0].internalMessageId, messageState.status)) {
                                this.#notifier.notify(SfuEvent.MESSAGE_STATE, messageState.status);
                            }
                        } else if (data[0].type === InternalApi.USER_LIST) {
                            const userList = data[0] as UserListEvent;
                            promises.resolve(data[0].internalMessageId, userList.list);
                            this.#notifier.notify(SfuEvent.USER_LIST, userList.list);
                        } else if (data[0].type === InternalApi.USER_CALENDAR) {
                            const calendar = data[0] as UserCalendarEvent;
                            promises.resolve(data[0].internalMessageId, calendar.calendar);
                            this.#notifier.notify(SfuEvent.USER_CALENDAR, calendar.calendar);
                        } else if (data[0].type === InternalApi.NEW_CHAT) {
                            const chatInfo = data[0] as NewChatEvent;
                            if(!promises.resolve(data[0].internalMessageId, chatInfo.info)) {
                                this.#notifier.notify(SfuEvent.NEW_CHAT, chatInfo.info);
                            }
                        } else if (data[0].type === InternalApi.CONTACT_INVITE) {
                            const invite = data[0] as ContactInviteEvent;
                            if(!promises.resolve(data[0].internalMessageId, invite.invite)) {
                                this.#notifier.notify(SfuEvent.CONTACT_INVITE, invite.invite);
                            }
                        } else if (data[0].type === InternalApi.CONTACT_UPDATED) {
                            const contact = data[0] as ContactUpdateEvent;
                            if(!promises.resolve(data[0].internalMessageId, contact.contact)) {
                                this.#notifier.notify(SfuEvent.CONTACT_UPDATE, contact.contact);
                            }
                        } else if (data[0].type === InternalApi.CONTACT_REMOVED) {
                            const contact = data[0] as ContactRemovedEvent;
                            if(!promises.resolve(data[0].internalMessageId, contact.contact)) {
                                this.#notifier.notify(SfuEvent.CONTACT_REMOVED, contact.contact);
                            }
                        } else if (data[0].type === InternalApi.CHAT_DELETED) {
                            const chat = data[0] as RemovedChatEvent;
                            if(!promises.resolve(data[0].internalMessageId, chat.info)) {
                                this.#notifier.notify(SfuEvent.CHAT_DELETED, chat.info);
                            }
                        } else if (data[0].type === InternalApi.CHAT_UPDATED) {
                            const chat = data[0] as UpdateChatEvent;
                            promises.resolve(data[0].internalMessageId, chat.info);
                            this.#notifier.notify(SfuEvent.CHAT_UPDATED, chat.info);
                        } else if (data[0].type === InternalApi.USER_CHATS) {
                            const chats = data[0] as ChatsEvent;
                            promises.resolve(data[0].internalMessageId, chats.chats);
                            this.#notifier.notify(SfuEvent.USER_CHATS, chats.chats);
                        } else if (data[0].type === InternalApi.PUBLIC_CHANNELS) {
                            const channels = data[0] as PublicChannelsEvent;
                            promises.resolve(data[0].internalMessageId, channels.channels);
                            this.#notifier.notify(SfuEvent.PUBLIC_CHANNELS, channels.channels);
                        } else if (data[0].type === InternalApi.CHAT_LOADED) {
                            const chat = data[0] as ChatLoadedEvent;
                            promises.resolve(data[0].internalMessageId, chat.chat);
                            this.#notifier.notify(SfuEvent.CHAT_LOADED, chat.chat);
                        } else if (data[0].type === RoomEvent.OPERATION_FAILED && promises.promised(data[0].internalMessageId)) {
                            promises.reject(data[0].internalMessageId, data[0] as OperationFailedEvent);
                        } else if (data[0].type === SfuEvent.ACK && promises.promised(data[0].internalMessageId)) {
                            promises.resolve(data[0].internalMessageId);
                        } else if (data[0].roomName && data[0].roomName.length > 0) {
                            //room event
                            const room = this.#rooms[data[0].roomName];
                            if (room) {
                                room.processEvent(data[0]);
                            }
                        } else {
                            this.#notifier.notify(data[0].type as SfuEvent, data[0]);
                        }
                        break;
                    case "failed":
                        this.#notifier.notify(SfuEvent.FAILED, data[0]);
                        break;
                }
            },
            (e) => {
                self.#_state = State.FAILED;
                self.#notifier.notify(SfuEvent.FAILED, e as InternalMessage);
            },
            () => {
                self.#_state = State.DISCONNECTED;
                self.disconnect();
                self.#notifier.notify(SfuEvent.DISCONNECTED);
            }
        );
    }

    #emmitAction(action: InternalApi, data: object, resolve: Function, reject: Function) {
        const id = uuidv4();
        promises.add(id, resolve, reject);
        this.#connection.send(action, {
            ...data,
            internalMessageId: id
        });
    }

    #checkAuthenticated() {
        if (this.#_state !== State.AUTHENTICATED) {
            throw new Error("User isn't authenticated, current state " + this.#_state);
        }
    }

    public sendMessage(msg: {
        body: string,
        chatId: string,
        attachments?: Array<MessageAttachment>
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<MessageStatus>(function(resolve, reject) {
            self.#checkAuthenticated();
            if (!msg) {
                reject(new Error("Can't send null message"));
            } else if ((!msg.body || msg.body === "") && (!msg.attachments || msg.attachments.length === 0)) {
                reject(new Error("Can't send message without content"));
            } else if (!msg.chatId || msg.chatId === "") {
                reject(new Error("Can't send message without a chatId"));
            } else {
                self.#emmitAction(InternalApi.SEND_MESSAGE, {
                    body: msg.body,
                    chatId: msg.chatId,
                    attachments: msg.attachments
                }, resolve, reject);
            }
        });
    };

    public markMessageRead(msg: {
        id: string,
        chatId: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<MessageStatus>(function(resolve, reject) {
            if (!msg) {
                reject(new Error("Can't mark null message"));
            } else if (!msg.chatId || msg.chatId === "") {
                reject(new Error("Can't mark message without a chatId"));
            } else if (!msg.id || msg.id === "") {
                reject(new Error("Can't mark message without massage id"));
            } else {
                self.#emmitAction(InternalApi.MARK_MESSAGE_READ, {
                    id: msg.id,
                    chatId: msg.chatId
                }, resolve, reject);
            }
        });
    }

    public markMessageUnread(msg: {
        id: string,
        chatId: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<MessageStatus>(function(resolve, reject) {
            if (!msg) {
                reject(new Error("Can't mark null message"));
            } else if (!msg.chatId || msg.chatId === "") {
                reject(new Error("Can't mark message without a chatId"));
            } else if (!msg.id || msg.id === "") {
                reject(new Error("Can't mark message without massage id"));
            } else {
                self.#emmitAction(InternalApi.MARK_MESSAGE_UNREAD, {
                    id: msg.id,
                    chatId: msg.chatId
                }, resolve, reject);
            }
        });
    }

    public getUserList() {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<Array<User>>(function (resolve, reject){
            self.#emmitAction(InternalApi.GET_USER_LIST, {}, resolve, reject);
        });
    };

    public getUserCalendar() {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<Calendar>(function (resolve, reject){
            self.#emmitAction(InternalApi.GET_USER_CALENDAR, {}, resolve, reject);
        });
    };

    public addCalendarEvent(event: {
        id: string,
        title: string,
        description: string,
        start: number,
        end: number,
        recurring: boolean,
        accessCode: string,
        waitingRoom: boolean,
        usePMI: boolean,
        ownerVideo: boolean,
        participantVideo: boolean
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<void>(function(resolve, reject){
            self.#emmitAction(InternalApi.ADD_CALENDAR_EVENT, {
                event: event
            }, resolve, reject);
        });
    };

    public removeCalendarEvent(event: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<void>(function(resolve, reject){
            self.#emmitAction(InternalApi.REMOVE_CALENDAR_EVENT, {
                event: event
            }, resolve, reject);
        });
    };

    public getUserChats() {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<ChatMap>(function (resolve, reject){
            self.#emmitAction(InternalApi.GET_USER_CHATS, {}, resolve, reject);
        });
    };

    public getPublicChannels() {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<ChatMap>(function (resolve, reject){
            self.#emmitAction(InternalApi.GET_PUBLIC_CHANNELS, {}, resolve, reject);
        });
    };

    public loadChat(chat: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<Chat>(function (resolve, reject){
            self.#emmitAction(InternalApi.LOAD_CHAT, {id: chat.id}, resolve, reject);
        });
    };

    public createChat(chat: {
        id?: string,
        name?: string,
        members?: Array<string>,
        favourite?: boolean,
        channel?: boolean,
        channelType?: ChannelType,
        channelSendPolicy?: ChannelSendPolicy,
        sendPermissionList?: Array<string>
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.CREATE_CHAT, {
                id: chat.id,
                name: chat.name,
                members: chat.members,
                favourite: chat.favourite,
                channel: chat.channel,
                channelType: chat.channelType,
                channelSendPolicy: chat.channelSendPolicy,
                sendPermissionList: chat.sendPermissionList
            }, resolve, reject);
        });
    };

    public deleteChat(chat: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<void>(function (resolve, reject){
            self.#emmitAction(InternalApi.DELETE_CHAT, {id: chat.id}, resolve, reject);
        });
    };

    public renameChat(chat: {
        id: string,
        name: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<void>(function (resolve, reject){
            self.#emmitAction(InternalApi.RENAME_CHAT, {id: chat.id, name: chat.name}, resolve, reject);
        });
    };

    public addMemberToChat(chat: {
        id: string,
        member: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.ADD_MEMBER_TO_CHAT, {id: chat.id, member: chat.member}, resolve, reject);
        });
    };

    public removeMemberFromChat(chat: {
        id: string,
        member: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.REMOVE_MEMBER_FROM_CHAT, {id: chat.id, member: chat.member}, resolve, reject);
        });
    };

    public inviteContact(invite: {
        to: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<User>(function (resolve, reject){
            self.#emmitAction(InternalApi.INVITE_CONTACT, {from: self.#_user.username, to: invite.to}, resolve, reject);
        });
    }

    public confirmContact(invite: {
        id: string,
        from: string,
        to: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<User>(function (resolve, reject){
            self.#emmitAction(InternalApi.CONFIRM_CONTACT, {from: invite.from, to: invite.to, id: invite.id}, resolve, reject);
        });
    }

    public removeContact(contact: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<User>(function (resolve, reject){
            self.#emmitAction(InternalApi.REMOVE_CONTACT, {id: contact.id}, resolve, reject);
        });
    }

    public updateChannelSendPolicy(channel: {
        id: string,
        channelSendPolicy: ChannelSendPolicy
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.UPDATE_CHANNEL_SEND_POLICY, {
                id: channel.id,
                channelSendPolicy: channel.channelSendPolicy
            }, resolve, reject);
        });
    }

    public addChannelSendPermissionListMember(channel: {
        id: string,
        member: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.ADD_CHANNEL_SEND_PERMISSION_LIST_MEMBER, {
                id: channel.id,
                member: channel.member
            }, resolve, reject);
        });
    }

    public removeChannelSendPermissionListMember(channel: {
        id: string,
        member: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.REMOVE_CHANNEL_SEND_PERMISSION_LIST_MEMBER, {
                id: channel.id,
                member: channel.member
            }, resolve, reject);
        });
    }

    public addChatToFavourites(chat: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.ADD_CHAT_TO_FAVOURITES, {id: chat.id}, resolve, reject);
        });
    }

    public removeChatFromFavourites(chat: {
        id: string
    }) {
        this.#checkAuthenticated();
        const self = this;
        return new Promise<UserSpecificChatInfo>(function (resolve, reject){
            self.#emmitAction(InternalApi.REMOVE_CHAT_FROM_FAVOURITES, {id: chat.id}, resolve, reject);
        });
    }

    public createRoom(options: {
        pc: RTCPeerConnection,
        name: string,
        pin: string
    }) {
        if (!options) {
            throw new TypeError("No options provided");
        }
        this.#checkAuthenticated();
        const opt = {
            connection: this.#connection,
            name: options.name,
            pin: options.pin,
            pc: options.pc
        };
        const room = new RoomExtended(this.#connection, options.pc, options.name, options.pin);
        this.#rooms[room.name()] = room;
        const self = this;
        const cleanup = () => {
            self.#rooms[room.name()].pc().close();
            self.#rooms[room.name()].pc().dispatchEvent(new Event("connectionstatechange"));
            delete self.#rooms[room.name()];
        };
        room.on(RoomEvent.LEFT, function(participant: LeftRoom){
            if (participant.name === self.#_user.nickname) {
                cleanup();
            }
        }).on(RoomEvent.EVICTED, function(participant: LeftRoom) {
            if (participant.name === self.#_user.nickname) {
                cleanup();
            }
        }).on(RoomEvent.FAILED, cleanup
        ).on(RoomEvent.OPERATION_FAILED, function(e: OperationFailedEvent) {
            if (Operations.ROOM_JOIN === e.operation) {
                cleanup();
            }
        });
        return room;
    };

    public getRoom(options: {
        name: string
    }) {
        if (!options) {
            throw new TypeError("No options provided");
        }
        this.#checkAuthenticated();
        return this.#rooms[options.name];
    }

    public disconnect() {
        for (const [key, value] of Object.entries(this.#rooms)) {
            value.leaveRoom();
        }
        this.#_user = undefined;
        this.#connection.close();
        this.#createConnection();
        this.#_state = State.DISCONNECTED;
        this.#rooms = {};
    };

    public user() {
        return this.#_user;
    }

    public server() {
        return this.#_server;
    }

    public state() {
        return this.#_state;
    }
    
    public on(event: SfuEvent, callback: (arg0: NotifyUnion) => void): SfuExtended {
        this.#notifier.add(event, callback);
        return this;
    };

    public off(event: SfuEvent, callback: (arg0: NotifyUnion) => void): SfuExtended {
        this.#notifier.remove(event, callback);
        return this;
    };

}