import {Connection} from "./connection";
import {InternalApi, InternalMessage, RoomState, SfuEvent, State} from "./constants";
import {Notifier} from "./notifier";
import {Room} from "./room";

export class Sfu {
    #connection: Connection;
    #url: string;
    #_nickname: string;
    #_state: State = State.NEW;
    #_room: Room;
    //TODO(naz): Provide union instead of InternalMessage
    #notifier: Notifier<SfuEvent, InternalMessage> = new Notifier<SfuEvent, InternalMessage>();

    public connect(options: {
        url: string,
        timeout?: number,
        nickname: string,
        logGroup: string
    }) {
        if (!options) {
            throw new TypeError("No options provided");
        }
        this.#url = options.url;
        this.#_nickname = options.nickname;
        const connectionConfig = {
            url: options.url,
            appName: InternalApi.P_APP,
            timeout: options.timeout ? options.timeout : 10000,
            custom: {
                nickname: options.nickname,
                //TODO(naz): rename to logGroup
                roomName: options.logGroup
            }
        };
        const self = this;
        return new Promise<void>(async (resolve, reject) => {
            self.#connection = new Connection(
                (name, data) =>  {
                    switch (name) {
                        case InternalApi.DEFAULT_METHOD:
                            if (data[0].roomName && data[0].roomName.length > 0) {
                                //room event
                                if (this.#_room) {
                                    this.#_room.processEvent(data[0]);
                                }
                            } else {
                                this.#notifier.notify(data[0].type as SfuEvent, data[0]);
                            }
                            break;
                        case "failed":
                            this.#notifier.notify(SfuEvent.FAILED, data[0]);
                            break;
                    }
                },
                (e) => {
                    self.#_state = State.FAILED;
                    self.#notifier.notify(SfuEvent.FAILED, e as InternalMessage);
                    reject(e);
                },
                () => {
                    self.#_state = State.DISCONNECTED;
                    self.#notifier.notify(SfuEvent.DISCONNECTED);
                }
            );
            await this.#connection.connect(connectionConfig);
            self.#_state = State.CONNECTED;
            this.#notifier.notify(SfuEvent.CONNECTED);
            resolve();
        });
    };

    public createRoom(options: {
        pc: RTCPeerConnection,
        name: string,
        pin: string
    }) {
        if (this.#_room) {
            switch (this.#_room.state()) {
                case RoomState.NEW:
                case RoomState.JOINED:
                case RoomState.PENDING:
                    throw new Error("Only one room at a time, already have one with state " + this.#_room.state());
                case RoomState.FAILED:
                case RoomState.DISPOSED:
                    this.#_room = new Room(this.#connection, options.pc, options.name, options.pin);
                    break;
                default:
                    throw new Error("Unknown room state " + this.#_room.state());
            }
        } else {
            this.#_room = new Room(this.#connection, options.pc, options.name, options.pin);
        }
        return this.#_room;
    }

    public disconnect() {
        if (this.#connection) {
            this.#connection.close();
        }
    }
    
    public room() {
        return this.#_room;
    }
    
    public nickname() {
        return this.#_nickname;
    }

    public state() {
        return this.#_state;
    }

    public on(event: SfuEvent, callback: (arg0: InternalMessage) => void): Sfu {
        this.#notifier.add(event, callback);
        return this;
    };

    public off(event: SfuEvent, callback: (arg0: InternalMessage) => void): Sfu {
        this.#notifier.remove(event, callback);
        return this;
    };

}