import {SfuExtended} from "../../src/sdk/sfu-extended";
import {
    ChannelSendPolicy,
    Invite,
    Message,
    MessageState,
    MessageStatus,
    SfuEvent,
    User,
    UserSpecificChatInfo
} from "../../src/sdk/constants";
import {
    CALENDAR_EVENT,
    TEST_MESSAGE_ATTACHMENT,
    TEST_PRIVATE_CHANNEL,
    TEST_PUBLIC_CHANNEL,
    TEST_USER_0,
    TEST_USER_1,
    url
} from "../util/constants";

async function connect(userConfig: {
    username: string,
    password: string,
    nickname: string
}) {
    const sfu = new SfuExtended();
    await sfu.connect({
        url: url,
        ...userConfig
    });
    return sfu;
}

describe("sfu-extended", () => {
    it("should connect", (done) => {
       const sfu = new SfuExtended();
       sfu.on(SfuEvent.CONNECTED, () => {
           const user = sfu.user();
           expect(user).toBeTruthy();
           expect(user.username).toEqual(TEST_USER_0.username);
           expect(user.nickname).toEqual(TEST_USER_0.nickname);
           expect(user.pmi).toBeTruthy();
           sfu.disconnect();
           done();
       });
       sfu.connect({
           url: url,
           ...TEST_USER_0
       });
    });
    it("should disconnect", (done) => {
        const sfu = new SfuExtended();
        sfu.on(SfuEvent.CONNECTED, () => {
            sfu.disconnect();
        }).on(SfuEvent.DISCONNECTED, () => {
            done();
        });
        sfu.connect({
            url: url,
            ...TEST_USER_0
        });
    });

    describe("contacts", () => {
        it("should load contacts", async () => {
            const sfu = await connect(TEST_USER_0);
            const list = await sfu.getUserList();
            expect(list.length).toBeGreaterThan(0);
            const user = list[0];
            expect(user.id).toBeTruthy();
            sfu.disconnect();
        });
        it("should invite contact", async () => {
            const sfu = await connect(TEST_USER_0);
            const list = await sfu.getUserList();
            const user = list[0];
            const invitedUser = await sfu.inviteContact({to: user.id});
            expect(invitedUser.invite).toBeTruthy();
            expect(invitedUser.invite.id).toBeTruthy();
            expect(invitedUser.invite.from).toEqual(TEST_USER_0.username);
            expect(invitedUser.invite.to).toEqual(user.id);
            await sfu.removeContact({id: user.id});
            sfu.disconnect();
        });
        it("should remove contact", async () => {
            const sfu = await connect(TEST_USER_0);
            const list = await sfu.getUserList();
            const user = list[0];
            const invitedUser = await sfu.inviteContact({to: user.id});
            expect(invitedUser.id).toEqual(user.id);
            expect(invitedUser.invite).toBeTruthy();
            const removedUser = await sfu.removeContact({id: user.id});
            expect(removedUser.id).toEqual(user.id);
            expect(removedUser.invite).toBeFalsy();
            const freshList = await sfu.getUserList();
            const freshUser = list[0];
            expect(freshUser.id).toEqual(user.id);
            expect(freshUser.invite).toBeFalsy();
            sfu.disconnect();
        });
        describe("notifications", () => {
            test("invited contact should receive invite", async (done) => {
                const sfu0 = await connect(TEST_USER_0);
                const sfu1 = await connect(TEST_USER_1);
                sfu1.on(SfuEvent.CONTACT_INVITE, async (msg) => {
                    const invite = msg as Invite;
                    await sfu0.removeContact({id: TEST_USER_1.username});
                    sfu0.disconnect();
                    sfu1.disconnect();
                    done();
                });
                await sfu0.inviteContact({to: TEST_USER_1.username});
            });
            test("invited contact should confirm invite", async (done) => {
                const sfu0 = await connect(TEST_USER_0);
                const sfu1 = await connect(TEST_USER_1);
                sfu1.on(SfuEvent.CONTACT_INVITE, async (msg) => {
                    const invite = msg as Invite;
                    sfu1.confirmContact(invite);
                });
                sfu0.on(SfuEvent.CONTACT_UPDATE, async (msg) => {
                    const user = msg as User;
                    expect(user.confirmed).toEqual(true);
                    expect(user.id).toEqual(TEST_USER_1.username);
                    await sfu0.removeContact(user);
                    sfu0.disconnect();
                    sfu1.disconnect();
                    done();
                })
                await sfu0.inviteContact({to: TEST_USER_1.username});
            });
        });
    });

    //TODO(naz): add/remove should return object, add should return id that was created at server side
    describe("calendar", () => {
        it("should load calendar", async () => {
            const sfu = await connect(TEST_USER_0);
            const calendar = await sfu.getUserCalendar();
            expect(calendar).toBeTruthy();
            expect(calendar.events).toBeTruthy();
            sfu.disconnect();
        });
        it("should add event", async () => {
            const sfu = await connect(TEST_USER_0);
            await sfu.addCalendarEvent(CALENDAR_EVENT);
            const calendar = await sfu.getUserCalendar();
            expect(calendar.events).toHaveProperty(CALENDAR_EVENT.id);
            await sfu.removeCalendarEvent(CALENDAR_EVENT);
            sfu.disconnect();
        });
        it("should remove event", async () => {
            const sfu = await connect(TEST_USER_0);
            await sfu.addCalendarEvent(CALENDAR_EVENT);
            let calendar = await sfu.getUserCalendar();
            expect(calendar.events).toHaveProperty(CALENDAR_EVENT.id);
            await sfu.removeCalendarEvent(CALENDAR_EVENT);
            calendar = await sfu.getUserCalendar();
            expect(calendar.events[CALENDAR_EVENT.id]).toBeFalsy();
            sfu.disconnect();
        });
    });

    describe("chat", () => {
        describe("chats", () => {
            it("should load chats", async () => {
                const sfu = await connect(TEST_USER_0);
                const chats = await sfu.getUserChats();
                expect(chats).toBeTruthy();
                sfu.disconnect();
            });
            it("should create chat", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                expect(chat).toBeTruthy();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            test("newly created chat without name should have it generated by server", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({
                    members: [
                        TEST_USER_0.username,
                        TEST_USER_1.username
                    ]
                });
                const expectedName = TEST_USER_0.nickname + ", " + TEST_USER_1.nickname;
                expect(chat.name).toEqual(expectedName);
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should delete chat", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                expect(chat).toBeTruthy();
                let chats = await sfu.getUserChats();
                expect(chats).toHaveProperty(chat.id);
                await sfu.deleteChat(chat);
                chats = await sfu.getUserChats();
                expect(chats[chat.id]).toBeFalsy();
                sfu.disconnect();
            });
            it("should add member", async () => {
                const sfu = await connect(TEST_USER_0);
                let chat = await sfu.createChat({});
                chat = await sfu.addMemberToChat({id: chat.id, member: TEST_USER_1.username});
                expect(chat.members).toContain(TEST_USER_1.username);
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            test("adding new member should change chat's name", async () => {
                const sfu = await connect(TEST_USER_0);
                let chat = await sfu.createChat({});
                chat = await sfu.addMemberToChat({id: chat.id, member: TEST_USER_1.username});
                const expectedName = TEST_USER_0.nickname + ", " + TEST_USER_1.nickname;
                expect(chat.name).toEqual(expectedName);
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should remove member", async () => {
                const sfu = await connect(TEST_USER_0);
                let chat = await sfu.createChat({});
                chat = await sfu.addMemberToChat({id: chat.id, member: TEST_USER_1.username});
                expect(chat.members).toContain(TEST_USER_1.username);
                chat = await sfu.removeMemberFromChat({id: chat.id, member: TEST_USER_1.username});
                const member = chat.members.find((member) => {member === TEST_USER_1.username});
                expect(member).toBeUndefined();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should send message", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                const status = await sfu.sendMessage({
                    chatId: chat.id,
                    body: "test message"
                });
                expect(status).toBeTruthy();
                expect(status.id).toBeTruthy();
                expect(status.date).toBeTruthy();
                expect(status.state).toBeTruthy();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should send message with attachment", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                const status = await sfu.sendMessage({
                    chatId: chat.id,
                    body: "test message",
                    attachments: [
                        TEST_MESSAGE_ATTACHMENT
                    ]
                });
                expect(status).toBeTruthy();
                expect(status.id).toBeTruthy();
                expect(status.date).toBeTruthy();
                expect(status.state).toBeTruthy();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should mark message as read", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                const status = await sfu.sendMessage({
                    chatId: chat.id,
                    body: "test message"
                });
                await sfu.markMessageRead(status);
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should mark message as unread", async () => {
                const sfu = await connect(TEST_USER_0);
                const chat = await sfu.createChat({});
                const status = await sfu.sendMessage({
                    chatId: chat.id,
                    body: "test message"
                });
                await sfu.markMessageUnread(status);
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            //TODO(naz): sfu.renameChat should return UserSpecificChatInfo with the new name
            it("should rename chat", async () => {
                const sfu = await connect(TEST_USER_0);
                const chatOriginal = await sfu.createChat({});
                const name = "chat_name";
                await sfu.renameChat({id: chatOriginal.id, name: name});
                const chats = await sfu.getUserChats();
                const chatRenamed = chats[chatOriginal.id];
                expect(chatRenamed.name).toEqual(name);
                await sfu.deleteChat(chatRenamed);
                sfu.disconnect();
            });
            test("chat's name should remain stable after rename was done", async () => {
                const sfu = await connect(TEST_USER_0);
                const chatOriginal = await sfu.createChat({});
                const name = "chat_name";
                await sfu.renameChat({id: chatOriginal.id, name: name});
                await sfu.addMemberToChat({id: chatOriginal.id, member: TEST_USER_1.username});
                const chats = await sfu.getUserChats();
                const chatRenamed = chats[chatOriginal.id];
                expect(chatRenamed.name).toEqual(name);
                await sfu.deleteChat(chatRenamed);
                sfu.disconnect();
            });
            it("should add chat to favourites", async () => {
                const sfu = await connect(TEST_USER_0);
                let chat = await sfu.createChat({});
                chat = await sfu.addChatToFavourites(chat);
                expect(chat.favourite).toBeTruthy();
                const chats = await sfu.getUserChats();
                chat = chats[chat.id];
                expect(chat.favourite).toBeTruthy();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            it("should remove chat from favourites", async () => {
                const sfu = await connect(TEST_USER_0);
                let chat = await sfu.createChat({});
                chat = await sfu.addChatToFavourites(chat);
                expect(chat.favourite).toBeTruthy();
                chat = await sfu.removeChatFromFavourites(chat);
                expect(chat.favourite).toBeFalsy();
                const chats = await sfu.getUserChats();
                chat = chats[chat.id];
                expect(chat.favourite).toBeFalsy();
                await sfu.deleteChat(chat);
                sfu.disconnect();
            });
            describe("notifications", () => {
                test("user should be notified when added to chat", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.NEW_CHAT, async (msg) => {
                        const chat1 = msg as UserSpecificChatInfo;
                        expect(chat1.id).toEqual(chat0.id);
                        await sfu0.deleteChat(chat0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let chat0 = await sfu0.createChat({});
                    await sfu0.addMemberToChat({id: chat0.id, member: TEST_USER_1.username});
                });
                test("user should be notified when removed from chat", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_DELETED, async (msg) => {
                        const chat1 = msg as UserSpecificChatInfo;
                        expect(chat1.id).toEqual(chat0.id);
                        await sfu0.deleteChat(chat0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    }).on(SfuEvent.NEW_CHAT, async (msg) => {
                        const chat1 = msg as UserSpecificChatInfo;
                        expect(chat1.id).toEqual(chat0.id);
                        await sfu0.removeMemberFromChat({id: chat0.id, member: TEST_USER_1.username});
                    });
                    let chat0 = await sfu0.createChat({});
                    await sfu0.addMemberToChat({id: chat0.id, member: TEST_USER_1.username});
                });
                test("user should be notified when chat deleted", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_DELETED, async (msg) => {
                        const chat1 = msg as UserSpecificChatInfo;
                        expect(chat1.id).toEqual(chat0.id);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    }).on(SfuEvent.NEW_CHAT, async (msg) => {
                        const chat1 = msg as UserSpecificChatInfo;
                        expect(chat1.id).toEqual(chat0.id);
                        await sfu0.deleteChat(chat0);
                    });
                    let chat0 = await sfu0.createChat({});
                    await sfu0.addMemberToChat({id: chat0.id, member: TEST_USER_1.username});
                });
                test("should receive message", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        await sfu0.deleteChat(chat0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let chat0 = await sfu0.createChat({
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    await sfu0.sendMessage({
                        chatId: chat0.id,
                        body: messageBody
                    });
                });
                test("should receive message with attachment", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        expect(message1.attachments.length).toEqual(1);
                        expect(message1.attachments[0].name).toEqual(TEST_MESSAGE_ATTACHMENT.name);
                        expect(message1.attachments[0].size).toEqual(TEST_MESSAGE_ATTACHMENT.size);
                        expect(message1.attachments[0].payload).toEqual(TEST_MESSAGE_ATTACHMENT.payload);
                        expect(message1.attachments[0].type).toEqual(TEST_MESSAGE_ATTACHMENT.type);
                        await sfu0.deleteChat(chat0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let chat0 = await sfu0.createChat({
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    const status = await sfu0.sendMessage({
                        chatId: chat0.id,
                        body: messageBody,
                        attachments: [
                            TEST_MESSAGE_ATTACHMENT
                        ]
                    });
                });
                test("user should be notified with message IM state", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu0.on(SfuEvent.MESSAGE_STATE, async (msg) => {
                        const state = msg as MessageStatus;
                        expect(state).toBeTruthy();
                        expect(state.state).toEqual(MessageState.FULL_DELIVERY_FULL_READ);
                        await sfu0.deleteChat(chat0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        await sfu1.markMessageRead(message1);
                    });
                    let chat0 = await sfu0.createChat({
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    await sfu0.sendMessage({
                        chatId: chat0.id,
                        body: messageBody
                    });
                });
            });
        });
        describe("channels", () => {
            it("should create public channel", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                expect(channel).toBeTruthy();
                expect(channel.id).toBeTruthy();
                expect(channel.channel).toBeTruthy();
                expect(channel.name).toBe(TEST_PUBLIC_CHANNEL.name);
                expect(channel.channelType).toBe(TEST_PUBLIC_CHANNEL.channelType);
                expect(channel.channelSendPolicy).toBe(TEST_PUBLIC_CHANNEL.channelSendPolicy);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should delete public channel", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                await sfu.deleteChat(channel);
                const pChannels = await sfu.getPublicChannels();
                expect(pChannels[channel.id]).toBeUndefined();
                sfu.disconnect();
            });
            it("should load public channels", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const pChannels = await  sfu.getPublicChannels();
                expect(pChannels).toHaveProperty(channel.id);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("Should load public channel user participates in into chats", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const chats = await  sfu.getUserChats();
                expect(chats).toHaveProperty(channel.id);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should create private channel", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PRIVATE_CHANNEL);
                expect(channel).toBeTruthy();
                expect(channel.id).toBeTruthy();
                expect(channel.channel).toBeTruthy();
                expect(channel.name).toBe(TEST_PRIVATE_CHANNEL.name);
                expect(channel.channelType).toBe(TEST_PRIVATE_CHANNEL.channelType);
                expect(channel.channelSendPolicy).toBe(TEST_PRIVATE_CHANNEL.channelSendPolicy);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should delete private channel", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PRIVATE_CHANNEL);
                await sfu.deleteChat(channel);
                const chats = await  sfu.getUserChats();
                expect(chats[channel.id]).toBeUndefined();
                sfu.disconnect();
            });
            it("should not load private channels into public channels", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PRIVATE_CHANNEL);
                const pChannels = await  sfu.getPublicChannels();
                expect(pChannels[channel.id]).toBeUndefined();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("Should load private channel into chats", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PRIVATE_CHANNEL);
                const chats = await  sfu.getUserChats();
                expect(chats).toHaveProperty(channel.id);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should add member", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                channel = await sfu.addMemberToChat({id: channel.id, member: TEST_USER_1.username});
                expect(channel.members).toContain(TEST_USER_1.username);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should remove member", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                channel = await sfu.addMemberToChat({id: channel.id, member: TEST_USER_1.username});
                expect(channel.members).toContain(TEST_USER_1.username);
                channel = await sfu.removeMemberFromChat({id: channel.id, member: TEST_USER_1.username});
                const member = channel.members.find((member) => {member === TEST_USER_1.username});
                expect(member).toBeUndefined();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should send message", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const status = await sfu.sendMessage({
                    chatId: channel.id,
                    body: "test message"
                });
                expect(status).toBeTruthy();
                expect(status.id).toBeTruthy();
                expect(status.date).toBeTruthy();
                expect(status.state).toBeTruthy();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should send message with attachment", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const status = await sfu.sendMessage({
                    chatId: channel.id,
                    body: "test message",
                    attachments: [
                        TEST_MESSAGE_ATTACHMENT
                    ]
                });
                expect(status).toBeTruthy();
                expect(status.id).toBeTruthy();
                expect(status.date).toBeTruthy();
                expect(status.state).toBeTruthy();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should mark message as read", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const status = await sfu.sendMessage({
                    chatId: channel.id,
                    body: "test message"
                });
                await sfu.markMessageRead(status);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should mark message as unread", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const status = await sfu.sendMessage({
                    chatId: channel.id,
                    body: "test message"
                });
                await sfu.markMessageUnread(status);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should rename channel", async () => {
                const sfu = await connect(TEST_USER_0);
                const channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                const name = "chat_name";
                await sfu.renameChat({id: channel.id, name: name});
                const pChannels = await sfu.getPublicChannels();
                const channelRenamed = pChannels[channel.id];
                expect(channelRenamed.name).toEqual(name);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should add channel to favourites", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                channel = await sfu.addChatToFavourites(channel);
                expect(channel.favourite).toBeTruthy();
                const chats = await sfu.getUserChats();
                channel = chats[channel.id];
                expect(channel.favourite).toBeTruthy();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should remove channel from favourites", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                channel = await sfu.addChatToFavourites(channel);
                expect(channel.favourite).toBeTruthy();
                channel = await sfu.removeChatFromFavourites(channel);
                expect(channel.favourite).toBeFalsy();
                const chats = await sfu.getUserChats();
                channel = chats[channel.id];
                expect(channel.favourite).toBeFalsy();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should change send policy", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat(TEST_PUBLIC_CHANNEL);
                expect(channel.channelSendPolicy).toBe(TEST_PUBLIC_CHANNEL.channelSendPolicy);
                const policy = ChannelSendPolicy.ADMIN_AND_LIST;
                channel = await sfu.updateChannelSendPolicy({id: channel.id, channelSendPolicy: policy});
                expect(channel.channelSendPolicy).toBe(policy);
                await sfu.deleteChat(channel);
            });
            it("should add member to policy list", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat({
                    ...TEST_PUBLIC_CHANNEL,
                    channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST
                });
                expect(channel.sendPermissionList.length).toBe(0);
                channel = await sfu.addChannelSendPermissionListMember({id: channel.id, member: TEST_USER_1.username});
                expect(channel.sendPermissionList).toContain(TEST_USER_1.username);
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            it("should remove member from policy list", async () => {
                const sfu = await connect(TEST_USER_0);
                let channel = await sfu.createChat({
                    ...TEST_PUBLIC_CHANNEL,
                    channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST
                });
                expect(channel.sendPermissionList.length).toBe(0);
                channel = await sfu.addChannelSendPermissionListMember({id: channel.id, member: TEST_USER_1.username});
                expect(channel.sendPermissionList).toContain(TEST_USER_1.username);
                channel = await sfu.removeChannelSendPermissionListMember({id: channel.id, member: TEST_USER_1.username});
                const member = channel.sendPermissionList.find((member) => member === TEST_USER_1.username);
                expect(member).toBeUndefined();
                await sfu.deleteChat(channel);
                sfu.disconnect();
            });
            describe("notifications", () => {
                test("user should be notified when added to channel", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.NEW_CHAT, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let channel0 = await sfu0.createChat(TEST_PUBLIC_CHANNEL);
                    await sfu0.addMemberToChat({id: channel0.id, member: TEST_USER_1.username});
                });
                test("user should be notified when removed from channel", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_DELETED, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    }).on(SfuEvent.NEW_CHAT, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        await sfu0.removeMemberFromChat({id: channel0.id, member: TEST_USER_1.username});
                    });
                    let channel0 = await sfu0.createChat(TEST_PUBLIC_CHANNEL);
                    await sfu0.addMemberToChat({id: channel0.id, member: TEST_USER_1.username});
                });
                test("user should be notified when channel deleted", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_DELETED, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    }).on(SfuEvent.NEW_CHAT, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        await sfu0.deleteChat(channel0);
                    });
                    let channel0 = await sfu0.createChat({});
                    await sfu0.addMemberToChat({id: channel0.id, member: TEST_USER_1.username});
                });
                test("should receive message", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    await sfu0.sendMessage({
                        chatId: channel0.id,
                        body: messageBody
                    });
                });
                test("should receive message with attachment", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        expect(message1.attachments.length).toEqual(1);
                        expect(message1.attachments[0].name).toEqual(TEST_MESSAGE_ATTACHMENT.name);
                        expect(message1.attachments[0].size).toEqual(TEST_MESSAGE_ATTACHMENT.size);
                        expect(message1.attachments[0].payload).toEqual(TEST_MESSAGE_ATTACHMENT.payload);
                        expect(message1.attachments[0].type).toEqual(TEST_MESSAGE_ATTACHMENT.type);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    const status = await sfu0.sendMessage({
                        chatId: channel0.id,
                        body: messageBody,
                        attachments: [
                            TEST_MESSAGE_ATTACHMENT
                        ]
                    });
                });
                test("user should be notified with message IM state", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu0.on(SfuEvent.MESSAGE_STATE, async (msg) => {
                        const state = msg as MessageStatus;
                        expect(state).toBeTruthy();
                        expect(state.state).toEqual(MessageState.FULL_DELIVERY_FULL_READ);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    sfu1.on(SfuEvent.MESSAGE, async (msg) => {
                        const message1 = msg as Message;
                        expect(message1.body).toEqual(messageBody);
                        await sfu1.markMessageRead(message1);
                    });
                    let channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        members: [TEST_USER_1.username]
                    });
                    const messageBody = "test message";
                    await sfu0.sendMessage({
                        chatId: channel0.id,
                        body: messageBody
                    });
                });
                test("member's state should update after it was added to send permission list", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_UPDATED, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        expect(channel1.sendPermissionList).toContain(TEST_USER_1.username);
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        members: [TEST_USER_1.username]
                    });
                    await sfu0.addChannelSendPermissionListMember({id: channel0.id, member: TEST_USER_1.username});
                });
                test("member's state should update after it was removed from send permission list", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    sfu1.on(SfuEvent.CHAT_UPDATED, async (msg) => {
                        const channel1 = msg as UserSpecificChatInfo;
                        expect(channel1.id).toEqual(channel0.id);
                        const member = channel1.sendPermissionList.find((member) => member === TEST_USER_1.username);
                        expect(member).toBeUndefined();
                        await sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    let channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST,
                        members: [TEST_USER_1.username],
                        sendPermissionList: [TEST_USER_1.username]
                    });
                    await sfu0.removeChannelSendPermissionListMember({id: channel0.id, member: TEST_USER_1.username});
                });
                test("public channel visible to everyone", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat(TEST_PUBLIC_CHANNEL);
                    const pChannels1 = await  sfu1.getPublicChannels();
                    expect(pChannels1).toHaveProperty(channel0.id);
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                    sfu1.disconnect();
                });
                test("private channel not visible to everyone", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat(TEST_PRIVATE_CHANNEL);
                    const pChannels1 = await  sfu1.getPublicChannels();
                    expect(pChannels1[channel0.id]).toBeUndefined();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                    sfu1.disconnect();
                });
                test("send policy EVERYONE", async (done) => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.EVERYONE,
                        members: [TEST_USER_1.username]
                    });
                    sfu0.on(SfuEvent.MESSAGE, (msg) => {
                        const message0 = msg as Message;
                        sfu0.deleteChat(channel0);
                        sfu0.disconnect();
                        sfu1.disconnect();
                        done();
                    });
                    await sfu1.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    });
                });
                test("send policy ADMIN should fail for non admin", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN,
                        members: [TEST_USER_1.username]
                    });
                    await expect(sfu1.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    })).rejects.toBeTruthy();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                    sfu1.disconnect();
                });
                test("send policy ADMIN should work for admin", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN,
                        members: [TEST_USER_1.username]
                    });
                    await expect(sfu0.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    })).resolves.toBeTruthy();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                });
                test("send policy ADMIN_AND_LIST should work for listed", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST,
                        members: [TEST_USER_1.username],
                        sendPermissionList: [TEST_USER_1.username]
                    });
                    await expect(sfu1.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    })).resolves.toBeTruthy();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                    sfu1.disconnect();
                });
                test("send policy ADMIN_AND_LIST should work for admin", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST,
                        members: [TEST_USER_1.username],
                        sendPermissionList: []
                    });
                    await expect(sfu0.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    })).resolves.toBeTruthy();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                });
                test("send policy ADMIN_AND_LIST should fail for unlisted", async () => {
                    const sfu0 = await connect(TEST_USER_0);
                    const sfu1 = await connect(TEST_USER_1);
                    const channel0 = await sfu0.createChat({
                        ...TEST_PUBLIC_CHANNEL,
                        channelSendPolicy: ChannelSendPolicy.ADMIN_AND_LIST,
                        members: [TEST_USER_1.username],
                        sendPermissionList: []
                    });
                    await expect(sfu1.sendMessage({
                        chatId: channel0.id,
                        body: "test message"
                    })).rejects.toBeTruthy();
                    await sfu0.deleteChat(channel0);
                    sfu0.disconnect();
                    sfu1.disconnect();
                });
            });
        });
    });
});